from haystack.utils.highlighting import Highlighter


class ExactHighlighter(Highlighter):
    '''
    This class customizes the Haystack Highlighter to allow for
    highlighting multi-word strings.
    https://django-haystack.readthedocs.io/en/master/highlighting.html#highlighter
    https://github.com/django-haystack/django-haystack/blob/master/haystack/utils/highlighting.py

    Use this class to build custom filters in `search_result.html`.
    See LA Metro as a model.
    '''
    def __init__(self, query, **kwargs):
        super(Highlighter, self).__init__()
        self.query_words = self.make_query_words(query)

    def make_query_words(self, query):
        query_words = set()
        if query.startswith('"') and query.endswith('"'):
            query_words.add(query.strip('\"'))
        else:
            query_words = set([word.lower() for word in query.split() if not word.startswith("-")])

        return query_words



# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-10-12 17:18
from __future__ import unicode_literals

import councilmatic_core.models
from django.db import migrations, models
import uuid

class Migration(migrations.Migration):

    dependencies = [
        ('councilmatic_core', '0011_auto_20161012_1204'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='bill',
            name='id',
        ),
        migrations.RemoveField(
            model_name='event',
            name='id',
        ),
        migrations.RemoveField(
            model_name='person',
            name='id',
        ),
        migrations.AlterField(
            model_name='bill',
            name='ocd_id',
            field=models.CharField(max_length=100, primary_key=True, serialize=False, unique=True),
        ),
        migrations.AlterField(
            model_name='event',
            name='ocd_id',
            field=models.CharField(max_length=100, primary_key=True, serialize=False, unique=True),
        ),
        migrations.AlterField(
            model_name='person',
            name='ocd_id',
            field=models.CharField(default=uuid.uuid4, max_length=100, primary_key=True, serialize=False, unique=True),
        ),
    ]



# Generated by Django 2.1.9 on 2019-06-11 21:34

import django.core.files.storage
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('legislative', '0008_longer_event_name'),
        ('councilmatic_core', '0045_auto_20190513_1250'),
    ]

    operations = [
        migrations.CreateModel(
            name='BillDocument',
            fields=[
                ('document', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='councilmatic_document', serialize=False, to='legislative.BillDocument')),
                ('full_text', models.TextField(blank=True, null=True)),
            ],
            options={
                'abstract': False,
            },
            bases=('legislative.billdocument',),
        ),
        migrations.AlterField(
            model_name='person',
            name='headshot',
            field=models.FileField(default='images/headshot_placeholder.png', storage=django.core.files.storage.FileSystemStorage(base_url='/', location='/Users/hannah/projects/chi-councilmatic/chicago/static'), upload_to='images'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.13 on 2017-09-14 20:43
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('councilmatic_core', '0032_auto_20170912_1301'),
    ]

    operations = [
        migrations.CreateModel(
            name='RelatedBill',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('related_bill_identifier', models.CharField(max_length=100)),
                ('central_bill', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='related_bills', to='councilmatic_core.Bill')),
            ],
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.13 on 2017-09-12 20:01
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('councilmatic_core', '0031_eventdocument_updated_at'),
    ]

    operations = [
        migrations.CreateModel(
            name='Subject',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('subject', models.CharField(max_length=255)),
            ],
        ),
        migrations.RemoveField(
            model_name='bill',
            name='subject',
        ),
        migrations.AlterField(
            model_name='eventdocument',
            name='updated_at',
            field=models.DateTimeField(default=None, null=True),
        ),
        migrations.AddField(
            model_name='subject',
            name='bill',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='subjects', to='councilmatic_core.Bill'),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('councilmatic_core', '0003_auto_20151214_1222'),
    ]

    operations = [
        migrations.AlterField(
            model_name='person',
            name='source_url',
            field=models.CharField(blank=True, max_length=255),
        ),
    ]



import datetime
import os

from django.db import models
from django.contrib.gis.db import models as geo_models
from django.core.exceptions import ImproperlyConfigured
from django.conf import settings
from django.urls import reverse, NoReverseMatch
from django.utils import timezone
from django.db.models import Case, When
from django.db.models.functions import Cast
from django.utils.functional import cached_property
from django.core.files.storage import FileSystemStorage

from proxy_overrides.related import ProxyForeignKey
import opencivicdata.legislative.models
import opencivicdata.core.models


static_storage = FileSystemStorage(location=os.path.join(settings.STATIC_PATH), base_url='/')

if not hasattr(settings, 'OCD_CITY_COUNCIL_NAME'):
    raise ImproperlyConfigured('You must define OCD_CITY_COUNCIL_NAME in settings.py')

MANUAL_HEADSHOTS = settings.MANUAL_HEADSHOTS if hasattr(settings, 'MANUAL_HEADSHOTS') else {}


class CastToDateTimeMixin:
    def cast_to_datetime(self, field):
        """
        Cast a given field from a CharField to a DateTimeField, converting empty
        strings to NULL in the process. Useful for CharFields that store timestamps
        as strings.
        """
        return Cast(
            Case(
                When(**{field: '', 'then': None}),
                default=field,
                output_field=models.CharField()
            ),
            models.DateTimeField()
        )


class PersonManager(models.Manager):
    def get_queryset(self, *args, **kwargs):
        from django.db.models import Prefetch

        qs = super().get_queryset(*args, **kwargs)

        return qs.prefetch_related(
            Prefetch('memberships', Membership.objects.filter(person__in=qs))
        )


class Person(opencivicdata.core.models.Person):
    objects = PersonManager()

    person = models.OneToOneField(opencivicdata.core.models.Person,
                                  on_delete=models.CASCADE,
                                  related_name='councilmatic_person',
                                  parent_link=True)

    headshot = models.FileField(upload_to='images/headshots',
                                storage=static_storage,
                                default='images/headshot_placeholder.png')

    slug = models.SlugField()

    def delete(self, **kwargs):
        kwargs['keep_parents'] = kwargs.get('keep_parents', True)
        super().delete(**kwargs)

    def __str__(self):
        return self.name

    @property
    def latest_council_seat(self):
        m = self.latest_council_membership
        if m and m.post:
            return m.post.label
        return ''

    def is_speaker(self):
        return True if self.memberships.filter(role='Speaker').first() else False

    @property
    def headshot_source(self):
        sources = self.sources.filter(url=self.headshot.url)
        if sources:
            return sources.get().note
        elif self.headshot:
            return settings.CITY_VOCAB['SOURCE']
        else:
            return None

    @property
    def primary_sponsorships(self):
        primary_sponsorships = self.billsponsorship_set.filter(primary=True)\
                                                       .prefetch_related('bill')

        def sponsorship_sort(sponsorship):
            '''
            Sponsorships of bills without recent action dates should appear last.
            '''
            return sponsorship.bill.last_action_date or datetime.datetime(datetime.MINYEAR, 1, 1)

        return sorted((s for s in primary_sponsorships), key=sponsorship_sort, reverse=True)

    @property
    def chair_role_memberships(self):
        if hasattr(settings, 'COMMITTEE_CHAIR_TITLE'):
            return self.memberships.filter(role=settings.COMMITTEE_CHAIR_TITLE).filter(end_date_dt__gt=timezone.now())
        else:
            return []

    @property
    def member_role_memberships(self):
        if hasattr(settings, 'COMMITTEE_MEMBER_TITLE'):
            return self.memberships.filter(role=settings.COMMITTEE_MEMBER_TITLE).filter(end_date_dt__gt=timezone.now())
        else:
            return []

    @property
    def latest_council_membership(self):
        filter_kwarg = {'organization__name': settings.OCD_CITY_COUNCIL_NAME}

        city_council_memberships = self.memberships.filter(**filter_kwarg)

        if city_council_memberships.count():
            return city_council_memberships.order_by('-start_date', '-end_date').first()

        return None

    @property
    def current_council_seat(self):
        m = self.latest_council_membership
        if m and m.end_date_dt > timezone.now():
            return m.post.label
        return ''

    @property
    def link_html(self):
        return "<a href='{}'>{}</a>".format(reverse('person', args=[self.slug]), self.name)


class Organization(opencivicdata.core.models.Organization):

    organization = models.OneToOneField(opencivicdata.core.models.Organization,
                                        on_delete=models.CASCADE,
                                        related_name='councilmatic_organization',
                                        parent_link=True)

    slug = models.SlugField(max_length=200)

    def delete(self, **kwargs):
        kwargs['keep_parents'] = kwargs.get('keep_parents', True)
        super().delete(**kwargs)

    def __str__(self):
        return self.name

    @classmethod
    def committees(cls):
        """
        grabs all organizations (1) classified as a committee & (2) with at least one member
        """
        return [o for o in cls.objects.filter(classification='committee')
                if any([m.end_date_dt > timezone.now() for m in o.memberships.all()])]

    @property
    def recent_activity(self):
        # setting arbitrary max of 300 b/c otherwise page will take forever to
        # load
        return self.actions.order_by('-date', '-bill__identifier', '-order')[:300]

    @property
    def recent_events(self):
        # need to look up event participants by name
        events = Event.objects.filter(participants__entity_type='organization', participants__name=self.name)
        events = events.order_by('-start_date').all()
        return events

    @property
    def upcoming_events(self):
        """
        grabs events in the future
        """
        # need to look up event participants by name
        events = Event.objects\
                      .filter(participants__entity_type='organization', participants__name=self.name)\
                      .filter(start_time__gt=timezone.now())\
                      .order_by('start_time')\
                      .all()
        return events

    @property
    def chairs(self):
        if hasattr(settings, 'COMMITTEE_CHAIR_TITLE'):
            chairs = self.memberships.filter(role=settings.COMMITTEE_CHAIR_TITLE).filter(end_date_dt__gt=timezone.now()).select_related('person__councilmatic_person')
            for chair in chairs:
                chair.person = chair.person.councilmatic_person
            return chairs
        else:
            return []

    @property
    def non_chair_members(self):
        if hasattr(settings, 'COMMITTEE_MEMBER_TITLE'):
            return self.memberships.filter(role=settings.COMMITTEE_MEMBER_TITLE).filter(end_date_dt__gt=timezone.now())
        else:
            return []

    @property
    def all_members(self):
        if hasattr(settings, 'COMMITTEE_MEMBER_TITLE'):
            return self.memberships.filter(end_date_dt__gt=timezone.now())
        else:
            return []

    @property
    def vice_chairs(self):
        if hasattr(settings, 'COMMITTEE_VICE_CHAIR_TITLE'):
            return self.memberships.filter(role=settings.COMMITTEE_VICE_CHAIR_TITLE).filter(end_date_dt__gt=timezone.now())
        else:
            return []

    @property
    def link_html(self):
        link_fmt = '<a href="{0}">{1}</a>'

        if self.classification == 'committee':

            try:
                link_path = reverse('{}:committee_detail'.format(settings.APP_NAME), args=(self.slug,))
            except NoReverseMatch:
                link_path = reverse('committee_detail', args=(self.slug,))

            return link_fmt.format(link_path, self.name)

        if self.classification == 'legislature':

            try:
                link_path = reverse('{}:council_members'.format(settings.APP_NAME))
            except NoReverseMatch:
                link_path = reverse('council_members')

            return link_fmt.format(link_path, self.name)

        return self.name


class Post(opencivicdata.core.models.Post):

    post = models.OneToOneField(opencivicdata.core.models.Post,
                                on_delete=models.CASCADE,
                                related_name='councilmatic_post',
                                parent_link=True)

    organization = ProxyForeignKey(
        Organization,
        related_name='posts',
        help_text="The Organization in which the post is held.",
        on_delete=models.CASCADE,
    )

    shape = geo_models.GeometryField(null=True)

    @cached_property
    def current_member(self):
        membership = self.memberships.filter(end_date_dt__gt=timezone.now())\
                                     .order_by('-end_date', '-start_date')\
                                     .select_related('person__councilmatic_person').first()
        if membership:
            membership.person = membership.person.councilmatic_person
            return membership


class MembershipManager(CastToDateTimeMixin, models.Manager):
    def get_queryset(self):
        return super().get_queryset().annotate(
            end_date_dt=self.cast_to_datetime('end_date'),
            start_date_dt=self.cast_to_datetime('start_date')
        )


class Membership(opencivicdata.core.models.Membership):
    class Meta:
        proxy = True

    objects = MembershipManager()

    organization = ProxyForeignKey(
        Organization,
        related_name='memberships',
        # memberships will go away if the org does
        on_delete=models.CASCADE,
        help_text="A link to the Organization in which the Person is a member."
    )

    person = ProxyForeignKey(
        Person,
        related_name='memberships',
        null=True,
        # Membership will just unlink if the person goes away
        on_delete=models.SET_NULL,
        help_text="A link to the Person that is a member of the Organization."
    )

    post = ProxyForeignKey(
        Post,
        related_name='memberships',
        null=True,
        # Membership will just unlink if the post goes away
        on_delete=models.SET_NULL,
        help_text="	The Post held by the member in the Organization."
    )


class EventManager(CastToDateTimeMixin, models.Manager):
    def get_queryset(self):
        return super().get_queryset().annotate(
            start_time=self.cast_to_datetime('start_date')
        )


class Event(opencivicdata.legislative.models.Event):

    event = models.OneToOneField(opencivicdata.legislative.models.Event,
                                 on_delete=models.CASCADE,
                                 related_name='councilmatic_event',
                                 parent_link=True)

    slug = models.SlugField(max_length=200)

    def delete(self, **kwargs):
        kwargs['keep_parents'] = kwargs.get('keep_parents', True)
        super().delete(**kwargs)

    objects = EventManager()

    @property
    def event_page_url(self):

        try:
            link = reverse('{}:event_detail'.format(settings.APP_NAME), args=(self.slug,))
        except NoReverseMatch:
            link = reverse('event_detail', args=(self.slug,))

        return link

    @property
    def link_html(self):
        return '<a href="{0}" title="View Event Details">{1}</a>'.format(self.event_page_url, self.name)

    @property
    def clean_agenda_items(self):
        agenda_items = self.agenda.order_by('order').all()
        agenda_deduped = []
        descriptions_seen = []
        for a in agenda_items:
            if a.description not in descriptions_seen:
                descriptions_seen.append(a.description)
                agenda_deduped.append(a)

        return agenda_deduped

    @classmethod
    def next_city_council_meeting(cls):
        if hasattr(settings, 'CITY_COUNCIL_MEETING_NAME'):
            return cls.objects.filter(name__icontains=settings.CITY_COUNCIL_MEETING_NAME)\
                              .filter(start_time__gt=timezone.now()).order_by('start_time').first()
        else:
            return None

    @classmethod
    def most_recent_past_city_council_meeting(cls):
        if hasattr(settings, 'CITY_COUNCIL_MEETING_NAME'):
            return cls.objects.filter(name__icontains=settings.CITY_COUNCIL_MEETING_NAME)\
                .filter(start_time__lt=timezone.now()).order_by('-start_time').first()
        else:
            return None

    @classmethod
    def upcoming_committee_meetings(cls):
        return cls.objects.filter(start_time__gt=timezone.now())\
                  .exclude(name__icontains=settings.CITY_COUNCIL_MEETING_NAME)\
                  .order_by('start_time').all()[:3]


class Bill(opencivicdata.legislative.models.Bill):

    bill = models.OneToOneField(opencivicdata.legislative.models.Bill,
                                on_delete=models.CASCADE,
                                related_name='councilmatic_bill',
                                parent_link=True)

    slug = models.SlugField()
    restrict_view = models.BooleanField(default=False)

    def delete(self, **kwargs):
        kwargs['keep_parents'] = kwargs.get('keep_parents', True)
        super().delete(**kwargs)

    def __str__(self):
        return self.friendly_name

    @property
    def bill_type(self):
        type = self.extras.get('local_classification')
        if not type:
            type, = self.classification

        return type

    @property
    def full_text(self):
        return self.extras.get('rtf_text')

    @property
    def ocr_full_text(self):
        return self.extras.get('plain_text')

    @property
    def html_text(self):
        return self.extras.get('html_text')

    @property
    def web_source(self):
        return self.sources.filter(note='web').get()

    @property
    def controlling_body(self):
        """
        grabs the organization that's currently 'responsible' for a bill
        """

        if self.current_action:
            related_orgs = self.current_action.related_entities.filter(
                entity_type='organization').all()
            # when a bill is referred from city council
            # to a committee, controlling body is the organization
            # the bill was referred to (a related org)
            if related_orgs:
                controlling_bodies = [rel.organization for rel in related_orgs]
                return controlling_bodies
            # otherwise, the controlling body is usually whatever organization
            # performed the most recent action (this is the case most of the
            # time)
            else:
                return [self.current_action.organization]
        else:
            return None

    @property
    def last_action_org(self):
        """
        grabs whatever organization performed the most recent action
        """
        return self.current_action.organization if self.current_action else None

    @property
    def ordered_actions(self):
        """
        returns all actions ordered by date in descending order
        """
        return self.actions.order_by('-order')

    @property
    def current_action(self):
        """
        grabs the most recent action on a bill
        """
        return self.actions.last()

    @property
    def last_action_date(self):
        if self.current_action:
            return self.current_action.date_dt

    @property
    def first_action(self):
        """
        grabs the first action on a bill
        """
        return self.actions.first() if self.actions.all() else None

    @property
    def date_passed(self):
        return self.actions\
                   .filter(classification='passage')\
                   .last().date_dt

    @property
    def friendly_name(self):
        """
        the bill title/headers displayed throughout the site (bill listings, bill detail pages)

        by default this returns the bill identifier - override this in
        custom subclass to construct a friendly name that makes sense locally
        """
        return self.identifier

    @property
    def primary_sponsor(self):
        """
        grabs the primary sponsorship for a bill
        """
        return self.sponsorships.filter(primary=True).first()

    @property
    def pseudo_topics(self):
        """
        returns a list of artificial topics for a bill, from the committees
        that have been involved in the bill's history (the actions)

        this serves as a backup when there isn't data on the real topics,
        so that bill listings can still have some useful tags populated
        """
        if self.actions.all():

            orgs = set([a.organization.name for a in self.actions.all() if
                        (a.organization.name != 'Mayor' and
                         a.organization.name != settings.CITY_COUNCIL_NAME)])

            if not orgs and self.controlling_body and \
                    self.controlling_body[0].name != settings.CITY_COUNCIL_NAME:

                orgs = self.controlling_body

            return list(orgs)
        else:
            return []

    @property
    def topics(self):
        """
        returns a list of topics for a bill

        override this in custom subclass for richer topic logic
        """
        return []

    @property
    def addresses(self):
        """
        returns a list of relevant addresses for a bill

        override this in custom subclass
        """
        return []

    @property
    def inferred_status(self):
        """
        infers bill status, to be displayed in colored labels next to bill names

        override this in custom subclass w/ richer logic for determining
        bill status, e.g. active, passed, approved, failed, stale
        """
        return None

    @property
    def listing_description(self):
        abstract = self.abstracts.first()
        if abstract:
            return abstract
        return self.title

    @property
    def full_text_doc_url(self):
        """
        override this if instead of having full text as string stored in
        full_text, it is a PDF document that you can embed on the page
        """
        return None

    @classmethod
    def bills_since(cls, date_cutoff):
        """
        grabs all bills that have had activity since a given date
        """
        return cls.objects.filter(last_action_date_dt__gte=date_cutoff)

    @classmethod
    def new_bills_since(cls, date_cutoff):
        """
        grabs all bills that have been added since a given date
        (bills_since = new_bills_since + updated_bills_since)
        """
        all_bills_since = cls.bills_since(date_cutoff)
        new_bills_since = [
            b for b in all_bills_since if b.first_action.date >= date_cutoff]
        return new_bills_since

    @classmethod
    def updated_bills_since(cls, date_cutoff):
        """
        grabs all previously existing bills that have had activity since a given date
        (bills_since = new_bills_since + updated_bills_since)
        """
        all_bills_since = cls.bills_since(date_cutoff)
        updated_bills_since = [
            b for b in all_bills_since if b.first_action.date < date_cutoff]
        return updated_bills_since

    @property
    def unique_related_upcoming_events(self):
        events = [r.agenda_item.event for r in self.eventrelatedentity_set.filter(
            agenda_item__event__start_date__gte=timezone.now()).all()]
        return list(set(events))


class BillSponsorship(opencivicdata.legislative.models.BillSponsorship):
    class Meta:
        proxy = True

    bill = ProxyForeignKey(Bill, related_name='sponsorships', on_delete=models.CASCADE)
    organization = ProxyForeignKey(Organization, null=True, on_delete=models.SET_NULL)
    person = ProxyForeignKey(Person, null=True, on_delete=models.SET_NULL)


class BillActionManager(CastToDateTimeMixin, models.Manager):
    def get_queryset(self):
        return super().get_queryset().annotate(
            date_dt=self.cast_to_datetime('date')
        )


class BillAction(opencivicdata.legislative.models.BillAction):
    class Meta:
        proxy = True

    objects = BillActionManager()

    bill = ProxyForeignKey(Bill,
                           related_name='actions',
                           on_delete=models.CASCADE)

    organization = ProxyForeignKey(Organization,
                                   related_name='actions',
                                   # don't let an org delete wipe out a bunch of bill actions
                                   on_delete=models.PROTECT)

    @property
    def label(self):
        c = self.classification
        if not c:
            return 'default'
        c = c[0]

        if c == 'committee-passage':
            return 'success'
        if c == 'passage':
            return 'success'
        if c == 'executive-signature':
            return 'success'
        if c == 'amendment-passage':
            return 'success'

        if c == 'amendment-introduction':
            return 'default'
        if c == 'introduction':
            return 'default'
        if c == 'committee-referral':
            return 'default'
        if c == 'filing':
            return 'default'
        if c == 'executive-received':
            return 'default'

        if c == 'deferred':
            return 'primary'

        else:
            return 'default'

    @cached_property
    def referred_org(self):
        if self.description == 'Referred':
            related_entity = self.related_entities.get()
            if related_entity:
                return related_entity.organization


class BillActionRelatedEntity(opencivicdata.legislative.models.BillActionRelatedEntity):
    class Meta:
        proxy = True

    action = ProxyForeignKey(BillAction,
                             related_name='related_entities',
                             on_delete=models.CASCADE)

    organization = ProxyForeignKey(Organization,
                                   null=True,
                                   on_delete=models.SET_NULL)


class BillDocument(opencivicdata.legislative.models.BillDocument):

    document = models.OneToOneField(opencivicdata.legislative.models.BillDocument,
                                    on_delete=models.CASCADE,
                                    related_name='councilmatic_document',
                                    parent_link=True)

    full_text = models.TextField(blank=True, null=True)



from django.db.models.signals import post_save, pre_delete
from django.dispatch import receiver
from django.utils.text import slugify, Truncator

from opencivicdata.core.models import (Organization as OCDOrganization,
                                       Person as OCDPerson,
                                       Post as OCDPost)
from opencivicdata.legislative.models import (Event as OCDEvent,
                                              Bill as OCDBill)

from councilmatic_core.models import (Organization as CouncilmaticOrganization,
                                      Person as CouncilmaticPerson,
                                      Event as CouncilmaticEvent,
                                      Bill as CouncilmaticBill,
                                      Post as CouncilmaticPost)


@receiver(post_save, sender=OCDOrganization)
def create_councilmatic_org(sender, instance, created, **kwargs):
    if created:
        ocd_part = instance.id.rsplit('-', 1)[-1]
        slug = '{0}-{1}'.format(slugify(instance.name), ocd_part)

        co = CouncilmaticOrganization(organization=instance,
                                      slug=slug)
        # just update the child table, not the parent table
        co.save_base(raw=True)


@receiver(post_save, sender=OCDPerson)
def create_councilmatic_person(sender, instance, created, **kwargs):
    if created:
        ocd_part = instance.id.rsplit('-', 1)[-1]
        slug = '{0}-{1}'.format(slugify(instance.name), ocd_part)

        cp = CouncilmaticPerson(person=instance,
                                slug=slug)
        # just update the child table, not the parent table
        cp.save_base(raw=True)

@receiver(post_save, sender=OCDEvent)
def create_councilmatic_event(sender, instance, created, **kwargs):
    if created:
        truncator = Truncator(instance.name)
        ocd_part = instance.id.rsplit('-', 1)[-1]
        slug = '{0}-{1}'.format(slugify(truncator.words(5)), ocd_part)

        ce = CouncilmaticEvent(event=instance,
                               slug=slug)
        # just update the child table, not the parent table
        ce.save_base(raw=True)

@receiver(post_save, sender=OCDBill)
def create_councilmatic_bill(sender, instance, created, **kwargs):
    if created:
        slug = slugify(instance.identifier)

        cb = CouncilmaticBill(bill=instance,
                              slug=slug)
        # just update the child table, not the parent table
        cb.save_base(raw=True)

        cb = CouncilmaticBill.objects.get(id=instance.id)
    else:
        cb = instance.councilmatic_bill

@receiver(post_save, sender=OCDPost)
def create_councilmatic_post(sender, instance, created, **kwargs):
    if created:
        cp = CouncilmaticPost(post=instance)
        cp.save_base(raw=True)



import os
import sys
import subprocess
import logging
import logging.config
import sqlalchemy as sa
import datetime
import signal
import json

from django.core.management.base import BaseCommand
from django.conf import settings
from django.db.models import Max, Q

from councilmatic_core.models import Bill

logging.config.dictConfig(settings.LOGGING)
logging.getLogger("requests").setLevel(logging.WARNING)
logger = logging.getLogger(__name__)

DB_CONN = 'postgresql://{USER}:{PASSWORD}@{HOST}:{PORT}/{NAME}'

engine = sa.create_engine(DB_CONN.format(**settings.DATABASES['default']),
                          convert_unicode=True,
                          server_side_cursors=True)

class Command(BaseCommand):
    help = 'Converts rtf-formatted legislative text to valid html'

    def add_arguments(self, parser):
        parser.add_argument(
            '--update_all',
            default=False,
            action='store_true',
            help='Update html_text in all bills.')

        parser.add_argument(
            '--update_empty',
            default=False,
            action='store_true',
            help='Update bills that currently do not have html_text.')

    def handle(self, *args, **options):
        self.connection = engine.connect()
        self.update_all = options['update_all']
        self.update_empty = options['update_empty']
        '''
        This command converts RTF from Legistar into valid HTML.
        The conversion employs "unoconv" - a CLI tool that imports and exports documents in LibreOffice. We run unoconv as a daemon process and kill it when the conversions finish.
        Three steps occur: (1) querying the database for bill full_text (i.e., the RTF from Legistar),
        (2) iteration over the query results, conversion to html, and creation of an inserts string,
        (3) updating the full_text field with new html.
        '''
        listener = subprocess.Popen(['unoconv', '--listener'], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        try:
            self.add_html()
        finally:
            listener.terminate()

    def get_rtf(self):
        self.connection.execute("SET local timezone to '{}'".format(settings.TIME_ZONE))
        '''
        The script, by default, only converts the most recently updated bills:
        the `max_updated` timestamp and the third (last) query option
        helps accomplish this.

        Specifically, the query determines which bills have been
        recently updated (or created) in the Councilmatic database,
        by looking for bills with an `updated_at` (i.e.,  max_updated)
        timestamp of equal or greater value.
        '''
        max_updated = Bill.objects.all().aggregate(Max('updated_at'))['updated_at__max']

        has_rtf_text = Q(extras__rtf_text__isnull=False)
        missing_html_text = Q(extras__html_text__isnull=True)
        after_max_update = Q(updated_at__gt=max_updated)

        if max_updated is None or self.update_all:
            qs = Bill.objects.filter(has_rtf_text)
        elif self.update_empty:
            qs = Bill.objects.filter(has_rtf_text & missing_html_text)
        else:
            qs = Bill.objects.filter(has_rtf_text & after_max_update)

        yield from qs

    def convert_rtf(self):
        rtf_results = self.get_rtf()

        logger.info('Converting RTF to HTML....')

        inserts = ''
        for bill in rtf_results:
            ocd_id = bill.id
            rtf_string = bill.extras['rtf_text']

            try:
                # For Python 3.4 and below
                process = subprocess.Popen(['unoconv', '--stdin', '--stdout', '-f', 'html'], preexec_fn=os.setsid, stdout=subprocess.PIPE, stdin=subprocess.PIPE, stderr=subprocess.DEVNULL)

                html_data, stderr_data = process.communicate(input=rtf_string.encode(), timeout=15)

                html = html_data.decode('utf-8')
            except subprocess.TimeoutExpired as e:
                os.killpg(os.getpgid(process.pid), signal.SIGTERM)

                logger.error(e)
                logger.error('Look at bill {}'.format(ocd_id))
                continue

            logger.info('Successful conversion of {}'.format(ocd_id))

            yield {'html': json.dumps(html), 'ocd_id': ocd_id}

    def add_html(self):
        html_results = self.convert_rtf()

        self.connection.execute("SET local timezone to '{}'".format(settings.TIME_ZONE))
        query = '''
            UPDATE opencivicdata_bill
            SET extras = jsonb_set(extras, '{html_text}', :html)
            WHERE id = :ocd_id
        '''

        chunk = []
        # for html, ocd_id in html_results:
        for bill_dict in html_results:
            chunk.append(bill_dict)
            if len(chunk) == 20:
                with self.connection.begin() as trans:
                    self.connection.execute(sa.text(query), chunk)

                    chunk = []

        # Update bills when less than 1,000 elements in a chunk.
        if chunk:
            with self.connection.begin() as trans:
                self.connection.execute(sa.text(query), chunk)

        logger.info('Bills have valid, viewable HTML!')



# Generated by Django 2.1.8 on 2019-05-13 17:50

import django.core.files.storage
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('legislative', '0008_longer_event_name'),
        ('core', '0004_auto_20171005_2028'),
        ('councilmatic_core', '0044_bill_restrict_view'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='action',
            name='_bill',
        ),
        migrations.RemoveField(
            model_name='action',
            name='_organization',
        ),
        migrations.RemoveField(
            model_name='actionrelatedentity',
            name='_action',
        ),
        migrations.RemoveField(
            model_name='billdocument',
            name='bill',
        ),
        migrations.RemoveField(
            model_name='eventagendaitem',
            name='bill',
        ),
        migrations.RemoveField(
            model_name='eventagendaitem',
            name='event',
        ),
        migrations.RemoveField(
            model_name='eventdocument',
            name='event',
        ),
        migrations.RemoveField(
            model_name='eventmedia',
            name='event',
        ),
        migrations.RemoveField(
            model_name='eventparticipant',
            name='event',
        ),
        migrations.RemoveField(
            model_name='membership',
            name='_organization',
        ),
        migrations.RemoveField(
            model_name='membership',
            name='_person',
        ),
        migrations.RemoveField(
            model_name='membership',
            name='_post',
        ),
        migrations.RemoveField(
            model_name='post',
            name='_organization',
        ),
        migrations.RemoveField(
            model_name='relatedbill',
            name='central_bill',
        ),
        migrations.RemoveField(
            model_name='sponsorship',
            name='_bill',
        ),
        migrations.RemoveField(
            model_name='sponsorship',
            name='_person',
        ),
        migrations.RemoveField(
            model_name='subject',
            name='bill',
        ),
        migrations.CreateModel(
            name='BillAction',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
            },
            bases=('legislative.billaction',),
        ),
        migrations.CreateModel(
            name='BillActionRelatedEntity',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
            },
            bases=('legislative.billactionrelatedentity',),
        ),
        migrations.CreateModel(
            name='BillSponsorship',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
            },
            bases=('legislative.billsponsorship',),
        ),
        migrations.RemoveField(
            model_name='bill',
            name='_from_organization',
        ),
        migrations.RemoveField(
            model_name='bill',
            name='_legislative_session',
        ),
        migrations.RemoveField(
            model_name='bill',
            name='abstract',
        ),
        migrations.RemoveField(
            model_name='bill',
            name='bill_type',
        ),
        migrations.RemoveField(
            model_name='bill',
            name='classification',
        ),
        migrations.RemoveField(
            model_name='bill',
            name='description',
        ),
        migrations.RemoveField(
            model_name='bill',
            name='full_text',
        ),
        migrations.RemoveField(
            model_name='bill',
            name='html_text',
        ),
        migrations.RemoveField(
            model_name='bill',
            name='identifier',
        ),
        migrations.RemoveField(
            model_name='bill',
            name='last_action_date',
        ),
        migrations.RemoveField(
            model_name='bill',
            name='ocd_created_at',
        ),
        migrations.RemoveField(
            model_name='bill',
            name='ocd_id',
        ),
        migrations.RemoveField(
            model_name='bill',
            name='ocd_updated_at',
        ),
        migrations.RemoveField(
            model_name='bill',
            name='ocr_full_text',
        ),
        migrations.RemoveField(
            model_name='bill',
            name='source_note',
        ),
        migrations.RemoveField(
            model_name='bill',
            name='source_url',
        ),
        migrations.RemoveField(
            model_name='bill',
            name='updated_at',
        ),
        migrations.RemoveField(
            model_name='event',
            name='all_day',
        ),
        migrations.RemoveField(
            model_name='event',
            name='classification',
        ),
        migrations.RemoveField(
            model_name='event',
            name='description',
        ),
        migrations.RemoveField(
            model_name='event',
            name='end_time',
        ),
        migrations.RemoveField(
            model_name='event',
            name='extras',
        ),
        migrations.RemoveField(
            model_name='event',
            name='location_name',
        ),
        migrations.RemoveField(
            model_name='event',
            name='location_url',
        ),
        migrations.RemoveField(
            model_name='event',
            name='name',
        ),
        migrations.RemoveField(
            model_name='event',
            name='ocd_created_at',
        ),
        migrations.RemoveField(
            model_name='event',
            name='ocd_id',
        ),
        migrations.RemoveField(
            model_name='event',
            name='ocd_updated_at',
        ),
        migrations.RemoveField(
            model_name='event',
            name='source_note',
        ),
        migrations.RemoveField(
            model_name='event',
            name='source_url',
        ),
        migrations.RemoveField(
            model_name='event',
            name='start_time',
        ),
        migrations.RemoveField(
            model_name='event',
            name='status',
        ),
        migrations.RemoveField(
            model_name='event',
            name='updated_at',
        ),
        migrations.RemoveField(
            model_name='organization',
            name='_parent',
        ),
        migrations.RemoveField(
            model_name='organization',
            name='classification',
        ),
        migrations.RemoveField(
            model_name='organization',
            name='jurisdiction',
        ),
        migrations.RemoveField(
            model_name='organization',
            name='name',
        ),
        migrations.RemoveField(
            model_name='organization',
            name='ocd_id',
        ),
        migrations.RemoveField(
            model_name='organization',
            name='source_url',
        ),
        migrations.RemoveField(
            model_name='organization',
            name='updated_at',
        ),
        migrations.RemoveField(
            model_name='person',
            name='email',
        ),
        migrations.RemoveField(
            model_name='person',
            name='name',
        ),
        migrations.RemoveField(
            model_name='person',
            name='ocd_id',
        ),
        migrations.RemoveField(
            model_name='person',
            name='source_note',
        ),
        migrations.RemoveField(
            model_name='person',
            name='source_url',
        ),
        migrations.RemoveField(
            model_name='person',
            name='updated_at',
        ),
        migrations.RemoveField(
            model_name='person',
            name='website_url',
        ),
        migrations.AddField(
            model_name='bill',
            name='bill',
            field=models.OneToOneField(default='foo', on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='councilmatic_bill', serialize=False, to='legislative.Bill'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='event',
            name='event',
            field=models.OneToOneField(default='foo', on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='councilmatic_event', serialize=False, to='legislative.Event'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='organization',
            name='organization',
            field=models.OneToOneField(default='foo', on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='councilmatic_organization', serialize=False, to='core.Organization'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='person',
            name='person',
            field=models.OneToOneField(default='foo', on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='councilmatic_person', serialize=False, to='core.Person'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='bill',
            name='slug',
            field=models.SlugField(),
        ),
        migrations.AlterField(
            model_name='event',
            name='slug',
            field=models.SlugField(max_length=200),
        ),
        migrations.AlterField(
            model_name='organization',
            name='slug',
            field=models.SlugField(max_length=200),
        ),
        migrations.AlterField(
            model_name='person',
            name='headshot',
            field=models.FileField(default='images/headshot_placeholder.png', storage=django.core.files.storage.FileSystemStorage(base_url='/', location='/Users/fgregg/work/chi-councilmatic/static'), upload_to='images'),
        ),
        migrations.AlterField(
            model_name='person',
            name='slug',
            field=models.SlugField(),
        ),
        migrations.DeleteModel(
            name='Action',
        ),
        migrations.DeleteModel(
            name='ActionRelatedEntity',
        ),
        migrations.DeleteModel(
            name='BillDocument',
        ),
        migrations.DeleteModel(
            name='EventAgendaItem',
        ),
        migrations.DeleteModel(
            name='EventDocument',
        ),
        migrations.DeleteModel(
            name='EventMedia',
        ),
        migrations.DeleteModel(
            name='EventParticipant',
        ),
        migrations.DeleteModel(
            name='Jurisdiction',
        ),
        migrations.DeleteModel(
            name='LegislativeSession',
        ),
        migrations.DeleteModel(
            name='Membership',
        ),
        migrations.DeleteModel(
            name='Post',
        ),
        migrations.DeleteModel(
            name='RelatedBill',
        ),
        migrations.DeleteModel(
            name='Sponsorship',
        ),
        migrations.DeleteModel(
            name='Subject',
        ),
        migrations.CreateModel(
            name='Membership',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
            },
            bases=('core.membership',),
        ),
        migrations.CreateModel(
            name='Post',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
            },
            bases=('core.post',),
        ),
    ]



import pysolr
import requests
import logging
import logging.config

from django.core.management.base import BaseCommand, CommandError
from django.conf import settings

from councilmatic_core.models import Bill


logging.config.dictConfig(settings.LOGGING)
logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("pysolr").setLevel(logging.WARNING)
logger = logging.getLogger(__name__)

class Command(BaseCommand):
    help = 'Checks for alignment between the Solr index and Councilmatic database'

    def handle(self, *args, **options):
        
        councilmatic_count = self.count_councilmatic_bills()

        try:
            solr_url = settings.HAYSTACK_CONNECTIONS['default']['URL']
            requests.get(solr_url)
        except requests.ConnectionError:
            message = "ConnectionError: Unable to connect to Solr at {} when running the data integrity check. Is Solr running?".format(solr_url)
            logger.error(message)
            raise Exception(message)

        solr = pysolr.Solr(solr_url)
        solr_index_count = solr.search(q='*:*').hits

        if solr_index_count != councilmatic_count:
            message = 'Solr index has {solr} entites. The Councilmatic database has {councilmatic} entities. That\'s a problem!'.format(solr=solr_index_count, councilmatic=councilmatic_count)
            logger.error(message)
            print('. . . . .\n')

            raise AssertionError(message)

        logger.info('Good news! Solr index has {solr} entites. The Councilmatic database has {councilmatic} entities.'.format(solr=solr_index_count, councilmatic=councilmatic_count))
        print('. . . . .\n')

    def count_councilmatic_bills(self):

        return Bill.objects.all().count()


from django.core.management import call_command
from django.core.management.base import BaseCommand, CommandError
from django.core.files import File
from django.conf import settings

from opencivicdata.core.models import Person as OCDPerson

import requests

class Command(BaseCommand):
    help = 'Attach headshots to councilmembers'

    def handle(self, *args, **opetions):
        for person in OCDPerson.objects.exclude(image=''):
            councilmatic_person = person.councilmatic_person
            filename = councilmatic_person.slug + '.jpg'
            response = requests.get(person.image)

            self.stdout.write('Downloading {}'.format(person.image))

            with open('/tmp/' + filename, 'wb') as f:
                f.write(response.content)

            with open('/tmp/' + filename, 'rb') as f:
                django_file = File(f)
                councilmatic_person.headshot.save(filename, django_file)

        self.stdout.write('Collecting static')

        call_command('collectstatic', verbosity=0, interactive=False)



from django import template
from django.template.defaultfilters import stringfilter
from django.utils.html import strip_tags
import re

from django.utils.safestring import mark_safe
from django.core.serializers import serialize
import json
from django.db.models.query import QuerySet
from urllib.parse import urlsplit, parse_qs, parse_qsl, urlencode

register = template.Library()


@register.filter
@stringfilter
def sentence_case(value):
    return value.replace("_", " ").capitalize()


@register.filter
@stringfilter
def inferred_status_label(status):
    return "<span class='label label-" + status.lower() + "'>" + status + "</span>"


@register.filter
@stringfilter
def facet_name(value):
    if value == 'bill_type':
        return 'Legislation type'
    if value == 'sponsorships':
        return 'Sponsor'
    if value == 'controlling_body':
        return 'Controlling body'
    if value == 'inferred_status':
        return 'Legislation status'


@register.filter
@stringfilter
def remove_action_subj(bill_action_desc):
    # removes 'by X' from bill action descriptions & expands abbrevs
    # for more readable action labels
    clean_action = re.sub(r'\bComm\b', 'Committee', bill_action_desc)
    clean_action = re.sub(r'\bRecved\b', 'Received', clean_action)
    clean_action = re.sub(r'[,\s]*by\s[^\s]*', '', clean_action)

    # shorten the really long action descriptions for approval w/ modifications
    if 'approved with modifications' in clean_action.lower():
        clean_action = 'Approved with Modifications'

    return clean_action


@register.filter
@stringfilter
def short_blurb(text_blob):
    if len(text_blob) > 196:
        blurb = text_blob[:196]
        blurb = blurb[:blurb.rfind(' ')] + ' ...'
        return blurb
    else:
        return text_blob


@register.filter
@stringfilter
def short_title(text_blob):
    if len(text_blob) > 28:
        blurb = text_blob[:24]
        blurb = blurb[:blurb.rfind(' ')] + ' ...'
        return blurb
    else:
        return text_blob


@register.filter
@stringfilter
def strip_mailto(email):
    return re.sub('mailto:', '', email)


@register.filter
@stringfilter
def committee_topic_only(committee_name):
    clean = re.sub('Committee on ', '', committee_name)
    clean = re.sub('Subcommittee on ', '', clean)
    if 'Mental Health, Developmental Disability' in clean:
        clean = 'Mental Health & Disability'
    return clean


@register.filter
@stringfilter
def clean_html(text):
    value = strip_tags(text).replace('\n', '')
    return re.sub(r'&(?:\w+|#\d+);', '', value)


@register.filter
@stringfilter
def alternative_identifiers(id_original):
    id_1 = re.sub(" ", " 0", id_original)
    id_2 = re.sub(" ", "", id_original)
    id_3 = re.sub(" ", "", id_1)
    return ' '.join(i for i in set([id_original, id_1, id_2, id_3]))


@register.filter
def get_item(dictionary, key):
    return dictionary.get(key)


@register.filter
def format_date_sort(s, fmt='%Y%m%d%H%M'):
    if s:
        return s.strftime(fmt)
    else:
        return '0'

'''
This tag updates the current URL params to reflect one of five sort options: date (asc, desc), title (asc, desc), and relevance.
The `order_by_filter.html` partial calls this tag. 
'''
@register.simple_tag
def search_with_querystring(request, **kwargs):
    query = urlsplit(request.get_full_path()).query
    query_as_dict = parse_qs(query)
    query_as_dict.update(kwargs)
   
    return '/search?' + urlencode(query_as_dict, doseq=True)


@register.simple_tag
def sort_direction(request):
    query_dict = request.GET
    return query_dict.get('order_by')


@register.filter
def reverse_sort(sort_direction):
    if sort_direction == 'asc':
        return 'desc'
    if sort_direction == 'desc':
        return 'asc'


@register.filter
def create_facet_string(selected_facets, query=None):
    facet_string = "/search/rss/?"

    if query:
        facet_string = "/search/rss/?q=" + query

    for key, values in selected_facets:
        for value in values:
            facet_string += "&selected_facets=" + key + ":" + value

    return facet_string

@register.filter
def remove_question(text):
    return text.rstrip('?')


# -*- coding: utf-8 -*-
# Generated by Django 1.9.13 on 2018-05-07 16:20
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('councilmatic_core', '0040_mediaevent_meta'),
    ]

    operations = [
        migrations.AddField(
            model_name='event',
            name='extras',
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        ),
        migrations.RunSQL('''
            UPDATE councilmatic_core_event
              SET extras = json_build_object('guid', guid)::jsonb
              WHERE guid IS NOT NULL
        ''', reverse_sql='''
            UPDATE councilmatic_core_event
              SET guid = extras->>'guid'
              WHERE extras->'guid' IS NOT NULL
        '''),
        migrations.RemoveField(
            model_name='event',
            name='guid',
        ),
    ]



import datetime
import itertools
import logging
import logging.config
import urllib.parse

from django.core.exceptions import ImproperlyConfigured
from django.core.management.base import BaseCommand
from django.conf import settings
import pytz

from opencivicdata.legislative.models import BillDocumentLink, EventDocumentLink


for configuration in ['AWS_KEY','AWS_SECRET']:
    if not hasattr(settings, configuration):
        raise ImproperlyConfigured(
            'Please define {0} in settings_deployment.py'.format(configuration))

logging.config.dictConfig(settings.LOGGING)
logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = 'Refreshes the property image cache by deleting documents that need to be newly created'

    def handle(self, *args, **options):
        from boto.s3.connection import S3Connection
        from boto.s3.key import Key
        from boto.exception import S3ResponseError

        document_urls = self._get_urls()
        aws_keys = self._create_keys(document_urls)

        s3_conn = S3Connection(settings.AWS_KEY, settings.AWS_SECRET)

        bucket = s3_conn.get_bucket('councilmatic-document-cache')

        bucket.delete_keys(aws_keys)

        logger.info(("Removed {} document(s) from the councilmatic-document-cache").format(len(aws_keys)))

    def _get_urls(self):
        app_timezone = pytz.timezone(settings.TIME_ZONE)
        one_hour_ago = app_timezone.localize(datetime.datetime.now()) - datetime.timedelta(hours=1)

        bill_docs = BillDocumentLink.objects.filter(document__bill__versions__isnull=False,
                                                    document__bill__updated_at__gte=one_hour_ago)\
                                            .values_list('url', flat=True)

        event_docs = EventDocumentLink.objects.filter(document__event__updated_at__gte=one_hour_ago)\
                                              .values_list('url', flat=True)

        return itertools.chain(bill_docs, event_docs)

    def _create_keys(self, document_urls):
        return [urllib.parse.quote_plus(url) for url in document_urls]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.13 on 2018-03-26 20:31
from __future__ import unicode_literals
import re

from django.conf import settings
from django.db import migrations, models
from django.utils.text import slugify

class Migration(migrations.Migration):
    def unmangle_identifier(apps, schema_editor):
        '''
        The `fix_bill_id` function mangled bill identifiers in at least two ways:
        (1) removing zeroes, e.g., 'Res 0229-2004' (original) becomes 'Res 229-2004' (mangled)
        (2) adding a space, e.g., 'T2018-1245' (original) becomes 'T 2018-1245' (mangled).

        For NYC, the second case ONLY affected bills that begin with 'T', and the first case did NOT affect bills that begin with 'T' - that is, NYC bills never follow these patterns 'T 0023-2015' or 'T0023-2015'.

        This migration unmangles identifiers and changes the slugs accordingly.
        '''

        Bill = apps.get_model('councilmatic_core', 'Bill')

        # New York City
        if settings.OCD_CITY_COUNCIL_NAME == 'New York City Council':
            deleted_zeroes = r'^((?!T\s)[A-Za-z]+)\s(\d{1,3})-([-\w]+)$'
            for bill in Bill.objects.filter(identifier__iregex=deleted_zeroes):
                match = re.match(deleted_zeroes, bill.identifier)
                unmangled_identifier = '{prefix} {mangled_count:0>4}-{remainder}'.format(prefix=match.group(1), 
                                                                                         mangled_count=match.group(2),
                                                                                         remainder=match.group(3))

                print('{} becomes {}'.format(bill.identifier, unmangled_identifier))
                try:
                    duplicate = Bill.objects.get(identifier=unmangled_identifier)
                    print('{} - duplicate found. Deleting.'.format(unmangled_identifier))
                    duplicate.delete()
                except Bill.DoesNotExist: 
                    pass

                bill.identifier = unmangled_identifier
                bill.slug = slugify(unmangled_identifier)
                bill.save()

            added_space = r'^(T)\s([-\d]+)$'
            for bill in Bill.objects.filter(identifier__iregex=added_space):
                match = re.match(added_space, bill.identifier)
                unmangled_identifier = '{mangled_prefix}{count}'.format(mangled_prefix=match.group(1), 
                                                                        count=match.group(2))
                
                print('{} becomes {}'.format(bill.identifier, unmangled_identifier))
                try:
                    duplicate = Bill.objects.get(identifier=unmangled_identifier)
                    print('{} - duplicate found. Deleting.'.format(unmangled_identifier))
                    duplicate.delete()
                except Bill.DoesNotExist: 
                    pass

                bill.identifier = unmangled_identifier
                bill.slug = slugify(unmangled_identifier)
                bill.save()
                
    dependencies = [
        ('councilmatic_core', '0036_auto_20180302_1247'),
    ]

    operations = [
        migrations.RunPython(unmangle_identifier)
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.13 on 2017-08-15 21:18
from __future__ import unicode_literals

import datetime
from django.db import migrations, models
from django.utils.timezone import utc


class Migration(migrations.Migration):

    dependencies = [
        ('councilmatic_core', '0030_eventagendaitem_notes'),
    ]

    operations = [
        migrations.AddField(
            model_name='eventdocument',
            name='updated_at',
            field=models.DateTimeField(default=None, null=True),
            preserve_default=False,
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.13 on 2019-02-07 14:17
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('councilmatic_core', '0043_membership_extras'),
    ]

    operations = [
        migrations.AddField(
            model_name='bill',
            name='restrict_view',
            field=models.BooleanField(default=False),
        ),
    ]



import re
import json
import itertools
from operator import attrgetter
import urllib
import requests
from dateutil.relativedelta import relativedelta
from dateutil import parser

from django.shortcuts import render, redirect
from django.conf import settings
from django.views.generic import TemplateView, ListView, DetailView
from django.views.decorators.clickjacking import xframe_options_exempt
from django.db.models import Max, Min
from django.core.cache import cache
from django.utils.text import slugify
from django.utils.decorators import method_decorator
from django.utils import timezone
from django.templatetags.static import static

from haystack.forms import FacetedSearchForm
from haystack.views import FacetedSearchView

from .models import Person, Bill, Organization, Event, Post


if (settings.USING_NOTIFICATIONS):
    from notifications.models import BillSearchSubscription


class CouncilmaticFacetedSearchView(FacetedSearchView):

    def extra_context(self):
        # Raise an error if Councilmatic cannot connect to Solr.
        # Most likely, Solr is down and needs restarting.
        try:
            solr_url = settings.HAYSTACK_CONNECTIONS['default']['URL']
            requests.get(solr_url)
        except requests.ConnectionError:
            raise Exception("ConnectionError: Unable to connect to Solr at {}. Is Solr running?".format(solr_url))

        extra = super(FacetedSearchView, self).extra_context()
        extra['request'] = self.request
        extra['facets'] = self.results.facet_counts()

        q_filters = ''

        url_params = [(p, val) for (p, val) in self.request.GET.items(
        ) if p != 'page' and p != 'selected_facets' and p != 'amp' and p != '_']
        selected_facet_vals = self.request.GET.getlist('selected_facets')
        search_term = self.request.GET.get('q')
        for facet_val in selected_facet_vals:
            url_params.append(('selected_facets', facet_val))
        if url_params:
            q_filters = urllib.parse.urlencode(url_params)

        extra['q_filters'] = q_filters

        selected_facets = {}
        for val in self.request.GET.getlist("selected_facets"):
            if val:
                [k, v] = val.split('_exact:', 1)
                try:
                    selected_facets[k].append(v)
                except KeyError:
                    selected_facets[k] = [v]

        extra['selected_facets'] = selected_facets

        extra['current_council_members'] = {
            p.current_member.person.name: p.label for p in Post.objects.all() if p.current_member
        }

        if (settings.USING_NOTIFICATIONS):
            extra['user_subscribed'] = False
            if self.request.user.is_authenticated:
                user = self.request.user
                extra['user'] = user

                search_params = {
                    'term': search_term,
                    'facets': selected_facets
                }

                try:
                    user.billsearchsubscriptions.get(user=user,
                                                     search_params__exact=search_params)
                    extra['user_subscribed'] = True
                except BillSearchSubscription.DoesNotExist:
                    extra['user_subscribed'] = False

        return extra


class CouncilmaticSearchForm(FacetedSearchForm):

    def __init__(self, *args, **kwargs):
        self.load_all = True

        super(CouncilmaticSearchForm, self).__init__(*args, **kwargs)

    def no_query_found(self):

        return self.searchqueryset.all()

# This is used by a context processor in settings.py to render these variables
# into the context of every page.


def city_context(request):
    relevant_settings = [
        'SITE_META',
        'FOOTER_CREDITS',
        'CITY_COUNCIL_NAME',
        'CITY_NAME',
        'CITY_NAME_SHORT',
        'CITY_VOCAB',
        'SEARCH_PLACEHOLDER_TEXT',
        'LEGISLATION_TYPE_DESCRIPTIONS',
        'LEGISTAR_URL',
        'DISQUS_SHORTNAME',
        'IMAGES',
        'MAP_CONFIG',
        'ANALYTICS_TRACKING_CODE',
        'ABOUT_BLURBS',
        'USING_NOTIFICATIONS',
        'GOOGLE_API_KEY'
    ]

    city_context = {s: getattr(settings, s, None) for s in relevant_settings}

    return city_context


class IndexView(TemplateView):

    template_name = 'councilmatic_core/index.html'
    bill_model = Bill
    event_model = Event

    def get_context_data(self, **kwargs):
        context = super(IndexView, self).get_context_data(**kwargs)

        recently_passed = self.find_recently_passed(self.bill_model)

        upcoming_meetings = list(
            self.event_model.upcoming_committee_meetings())

        context.update({
            'recently_passed': recently_passed,
            'next_council_meeting': self.event_model.next_city_council_meeting(),
            'upcoming_committee_meetings': upcoming_meetings,
        })

        return context

    def find_recently_passed(self, bill_model):
        recently_passed = []
        # Do you want to show recently passed bills? CUSTOMIZE FOR COUNCILMATIC.
        # Otherwise avoid any unnecessary iterative processes when loading the index.
        # Example: https://github.com/datamade/django-councilmatic/blob/154028d1bc4639fa1ec1af75139cab544abdd315/councilmatic_core/views.py#L139

        return recently_passed


class AboutView(TemplateView):
    template_name = 'councilmatic_core/about.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        context['timestamp'] = timezone.now().strftime('%m%d%Y%s')

        return context


class CouncilMembersView(ListView):
    template_name = 'councilmatic_core/council_members.html'
    context_object_name = 'posts'

    def map(self):
        map_geojson = {
            'type': 'FeatureCollection',
            'features': []
        }

        for post in self.object_list:
            if post.shape:

                council_member = "Vacant"
                detail_link = ""
                if post.current_member:
                    council_member = post.current_member.person.name
                    detail_link = post.current_member.person.slug

                feature = {
                    'type': 'Feature',
                    'geometry': json.loads(post.shape.json),
                    'properties': {
                        'district': post.label,
                        'council_member': council_member,
                        'detail_link': '/person/' + detail_link,
                        'select_id': 'polygon-{}'.format(slugify(post.label)),
                    }
                }

                map_geojson['features'].append(feature)

        return json.dumps(map_geojson)

    def get_queryset(self):
        get_kwarg = {'name': settings.OCD_CITY_COUNCIL_NAME}

        return Organization.objects.get(**get_kwarg).posts.all()

    def get_context_data(self, *args, **kwargs):
        context = super(CouncilMembersView, self).get_context_data(**kwargs)
        context['seo'] = self.get_seo_blob()

        if settings.MAP_CONFIG:
            context['map_geojson'] = self.map
        else:
            context['map_geojson'] = None

        return context

    def get_seo_blob(self):
        seo = {}
        seo.update(settings.SITE_META)
        return seo


class BillDetailView(DetailView):
    model = Bill
    template_name = 'councilmatic_core/legislation.html'
    context_object_name = 'legislation'

    def get_context_data(self, **kwargs):
        context = super(BillDetailView, self).get_context_data(**kwargs)

        context['actions'] = self.get_object().actions.all().order_by('-order')
        bill = context['legislation']

        seo = {}
        seo.update(settings.SITE_META)
        seo['site_desc'] = bill.listing_description
        seo['title'] = '%s - %s' % (bill.friendly_name,
                                    settings.SITE_META['site_name'])
        context['seo'] = seo

        context['user_subscribed'] = False
        if self.request.user.is_authenticated:
            user = self.request.user
            context['user'] = user
            # check if person of interest is subscribed to by user

            if settings.USING_NOTIFICATIONS:
                for bas in user.billactionsubscriptions.all():

                    if bill == bas.bill:
                        context['user_subscribed'] = True
                        break

        return context


@method_decorator(xframe_options_exempt, name='dispatch')
class BillWidgetView(BillDetailView):
    template_name = 'councilmatic_core/widgets/legislation.html'


class CommitteesView(ListView):
    template_name = 'councilmatic_core/committees.html'
    context_object_name = 'committees'

    def get_queryset(self):
        return Organization.committees


class CommitteeDetailView(DetailView):
    model = Organization
    template_name = 'councilmatic_core/committee.html'
    context_object_name = 'committee'

    def get_context_data(self, **kwargs):
        context = super(CommitteeDetailView, self).get_context_data(**kwargs)

        committee = context['committee']
        context['memberships'] = committee.memberships.all()

        description = None

        if getattr(settings, 'COMMITTEE_DESCRIPTIONS', None):
            description = settings.COMMITTEE_DESCRIPTIONS.get(committee.slug)
            context['committee_description'] = description

        seo = {}
        seo.update(settings.SITE_META)

        if description:
            seo['site_desc'] = description
        else:
            seo['site_desc'] = "See what %s's %s has been up to!" % (
                settings.CITY_COUNCIL_NAME, committee.name)

        seo['title'] = '%s - %s' % (committee.name,
                                    settings.SITE_META['site_name'])
        context['seo'] = seo

        context['user_subscribed_actions'] = False
        context['user_subscribed_events'] = False
        if self.request.user.is_authenticated:
            user = self.request.user
            context['user'] = user
            # check if person of interest is subscribed to by user

            if settings.USING_NOTIFICATIONS:
                for cas in user.committeeactionsubscriptions.all():
                    if committee == cas.committee:
                        context['user_subscribed_actions'] = True
                for ces in user.committeeeventsubscriptions.all():
                    if committee == ces.committee:
                        context['user_subscribed_events'] = True

        return context


@method_decorator(xframe_options_exempt, name='dispatch')
class CommitteeWidgetView(CommitteeDetailView):
    template_name = 'councilmatic_core/widgets/committee.html'


class PersonDetailView(DetailView):
    model = Person
    template_name = 'councilmatic_core/person.html'
    context_object_name = 'person'

    def get_context_data(self, **kwargs):
        context = super(PersonDetailView, self).get_context_data(**kwargs)

        person = context['person']
        context['sponsored_legislation'] = Bill.objects.filter(sponsorships__person=person)\
                                                       .filter(sponsorships__primary=True)\
                                                       .annotate(last_action=Max('actions__date'))\
                                                       .order_by('-last_action')[:10]

        title = ''
        if person.current_council_seat:
            title = '%s %s' % (person.current_council_seat,
                               settings.CITY_VOCAB['COUNCIL_MEMBER'])
        elif person.latest_council_seat:
            title = 'Former %s, %s' % (
                settings.CITY_VOCAB['COUNCIL_MEMBER'], person.latest_council_seat)
        elif getattr(settings, 'EXTRA_TITLES', None) and person.slug in settings.EXTRA_TITLES:
            title = settings.EXTRA_TITLES[person.slug]
        context['title'] = title

        seo = {}
        seo.update(settings.SITE_META)
        if person.current_council_seat:
            short_name = re.sub(r',.*', '', person.name)
            seo['site_desc'] = '%s - %s representative in %s. See what %s has been up to!' % (
                person.name, person.current_council_seat, settings.CITY_COUNCIL_NAME, short_name)
        else:
            seo['site_desc'] = 'Details on %s, %s' % (
                person.name, settings.CITY_COUNCIL_NAME)
        seo['title'] = '%s - %s' % (person.name,
                                    settings.SITE_META['site_name'])
        seo['image'] = static(person.headshot.url)
        context['seo'] = seo

        context['map_geojson'] = None

        if settings.MAP_CONFIG and person.latest_council_membership and person.latest_council_membership.post and person.latest_council_membership.post.shape:
            map_geojson = {
                'type': 'FeatureCollection',
                'features': []
            }

            feature = {
                'type': 'Feature',
                'geometry': json.loads(person.latest_council_membership.post.shape.json),
                'properties': {
                    'district': person.latest_council_membership.post.label,
                }
            }

            map_geojson['features'].append(feature)

            context['map_geojson'] = json.dumps(map_geojson)

        context['user_subscribed'] = False

        if (settings.USING_NOTIFICATIONS):

            if self.request.user.is_authenticated:
                user = self.request.user
                context['user'] = user
                # check if person of interest is subscribed to by user

                for ps in user.personsubscriptions.all():
                    if person == ps.person:
                        context['user_subscribed'] = True
                        break

        return context


@method_decorator(xframe_options_exempt, name='dispatch')
class PersonWidgetView(PersonDetailView):
    template_name = 'councilmatic_core/widgets/person.html'


class EventsView(ListView):
    template_name = 'councilmatic_core/events.html'

    def get_queryset(self):
        # Realize this is stupid. The reason this exists is so that
        # we can have this be a ListView in the inherited subclasses
        # if needed
        return []

    def get_context_data(self, **kwargs):
        context = super(EventsView, self).get_context_data(**kwargs)

        # Get year range for datepicker.
        aggregates = Event.objects.aggregate(
            Min('start_time'), Max('start_time'))

        context['year_range_min'] = aggregates['start_time__min'].year
        context['year_range_max'] = aggregates['start_time__max'].year

        # Did the user set date boundaries?
        date_str = self.request.GET.get('form_datetime')
        day_grouper = lambda x: x.start_time.date
        context['select_date'] = ''

        # If yes, then filter for dates.
        if date_str:
            context['date'] = date_str
            date_time = parser.parse(date_str)

            select_events = Event.objects.filter(start_time__gt=date_time)\
                .filter(start_time__lt=(date_time + relativedelta(months=1)))\
                .order_by('start_time')

            org_select_events = []

            for event_date, events in itertools.groupby(select_events, key=day_grouper):
                events = sorted(events, key=attrgetter('start_time'))
                org_select_events.append([event_date, events])

            context['select_events'] = org_select_events
            context['select_date'] = date_time.strftime("%B") + " " + date_time.strftime("%Y")

        # If no, then return upcoming events.
        else:
            # Upcoming events for the current month.
            upcoming_events = Event.objects.filter(start_time__gt=timezone.now())\
                .filter(start_time__lt=(timezone.now() + relativedelta(months=1)))\
                .order_by('start_time')

            if len(upcoming_events) < 3:
                # Upcoming events for the next month, plus two or three from previous months.
                upcoming_events = Event.objects.filter(start_time__gt=timezone.now())\
                    .filter(start_time__lt=(timezone.now() + relativedelta(months=2)))\
                    .order_by('start_time')

            org_upcoming_events = []

            for event_date, events in itertools.groupby(upcoming_events, key=day_grouper):
                events = sorted(events, key=attrgetter('start_time'))
                org_upcoming_events.append([event_date, events])

            context['upcoming_events'] = org_upcoming_events

        context['user_subscribed'] = False
        if self.request.user.is_authenticated:
            user = self.request.user
            context['user'] = user

            if settings.USING_NOTIFICATIONS:
                if (len(user.eventssubscriptions.all()) > 0):
                    context['user_subscribed'] = True

        return context


class EventDetailView(DetailView):
    template_name = 'councilmatic_core/event.html'
    model = Event
    context_object_name = 'event'

    def get_context_data(self, **kwargs):
        context = super(EventDetailView, self).get_context_data(**kwargs)
        event = context['event']

        participants = [p.entity_name for p in event.participants.all()]
        context['participants'] = Organization.objects.filter(
            name__in=participants)

        seo = {}
        seo.update(settings.SITE_META)
        seo['site_desc'] = 'Public city council event on %s/%s/%s - view event participants & agenda items' % (
            event.start_time.month, event.start_time.day, event.start_time.year)
        seo['title'] = '%s Event - %s' % (event.name,
                                          settings.SITE_META['site_name'])
        context['seo'] = seo

        return context


def flush(request, flush_key):

    try:
        if flush_key == settings.FLUSH_KEY:
            cache.clear()
    except AttributeError:
        print("\n\n** NOTE: to use flush-cache, set FLUSH_KEY in settings_local.py **\n\n")

    return redirect('index')


@xframe_options_exempt
def pdfviewer(request):
    return render(request, 'councilmatic_core/pdfviewer.html')



# Generated by Django 2.1.9 on 2019-06-18 17:30
import os

from django.conf import settings
import django.core.files.storage
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('councilmatic_core', '0046_subclass_ocd_billdocument'),
    ]

    operations = [
        migrations.AlterField(
            model_name='person',
            name='headshot',
            field=models.FileField(default='images/headshot_placeholder.png', storage=django.core.files.storage.FileSystemStorage(base_url='/', location=os.path.join(settings.STATIC_PATH)), upload_to='images/headshots'),
        ),
    ]
